"""# Useful stuff for hikari-based discord bot projects.

```
>>> from nya.{feature_module} import feature
```

### Nyaaa :3

```
>>> from nya.url import make_loopback_link
>>> make_loopback_link(ctx, "link text", tooltip="tooltip") # nya :3
'[link text](https://discord.com/channels/507642999992352779/507642999992352779 "tooltip")'
"""

from tcrutils._version import __version__

__tcr_rainbow__ = True
